--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------- 

-- Purpose: Flush the write input FIFO in io_ddr when not doing a write access.
-- Description:
--   The write input FIFO is flushed by when the io_ddr is not doing a write
--   access, so when idle or also when doing a read access.
--   When the io_ddr is starting a new write access, then the write input FIFO
--   gets filled. The filling starts dependent on:
--
--   . g_mode = "VAL" : immediately start filling on next valid data
--   . g_mode = "SOP" : start filling on next sop
--   . g_mode = "SYN" : start filling on next sync
--
--   . g_use_channel = TRUE : start filling when channel matches g_start_channel
--
--  Remarks:
--  . The flushing does ensure that the write FIFO does not run full. If the
--    write input FIFO is a mixed width FIFO with narrow write data, then it
--    may not be possible to read the FIFO empty, because a wide data word
--    can only be read when it is complete. Typically this behaviour is fine
--    in applications, so no need to try to somehow flush an incomplete last
--    wide word from the FIFO.

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

ENTITY io_ddr_driver_flush_ctrl IS 
  GENERIC (
    g_mode          : STRING := "VAL";  -- "VAL", "SOP", "SYN"
    g_use_channel   : BOOLEAN := FALSE;
    g_start_channel : NATURAL := 0;
    g_nof_channels  : POSITIVE := 1
   );  
  PORT ( 
    clk                : IN  STD_LOGIC;
    rst                : IN  STD_LOGIC;

    dvr_en             : IN  STD_LOGIC := '1';
    dvr_done           : IN  STD_LOGIC; 
    dvr_wr_flush_en    : IN  STD_LOGIC := '1';
    dvr_wr_not_rd      : IN  STD_LOGIC;       

    ctlr_wr_sosi       : IN  t_dp_sosi;

    ctlr_wr_flush_en   : OUT STD_LOGIC
   );
END io_ddr_driver_flush_ctrl;


ARCHITECTURE str OF io_ddr_driver_flush_ctrl IS

  CONSTANT c_channel_w     : NATURAL := ceil_log2(g_nof_channels);

  TYPE t_state IS (s_idle, s_flush, s_stop);

  SIGNAL state     : t_state;
  SIGNAL nxt_state : t_state;

  SIGNAL channel   : NATURAL RANGE 0 TO g_nof_channels-1;
  
  SIGNAL flush_dis : STD_LOGIC;

BEGIN

  -- Flush disable control
  no_channel: IF g_use_channel=FALSE GENERATE
    gen_valid : IF g_mode="VAL" GENERATE flush_dis <= ctlr_wr_sosi.valid; END GENERATE;
    gen_sop   : IF g_mode="SOP" GENERATE flush_dis <= ctlr_wr_sosi.sop  ; END GENERATE;
    gen_sync  : IF g_mode="SYN" GENERATE flush_dis <= ctlr_wr_sosi.sync ; END GENERATE;
  END GENERATE;
    
  use_channel: IF g_use_channel=TRUE GENERATE
    channel <= TO_UINT(ctlr_wr_sosi.channel(c_channel_w-1 DOWNTO 0));
    
    gen_valid : IF g_mode="VAL" GENERATE flush_dis <= '1' WHEN ctlr_wr_sosi.valid='1' AND channel=g_start_channel ELSE '0'; END GENERATE;
    gen_sop   : IF g_mode="SOP" GENERATE flush_dis <= '1' WHEN ctlr_wr_sosi.sop  ='1' AND channel=g_start_channel ELSE '0'; END GENERATE;
    gen_sync  : IF g_mode="SYN" GENERATE flush_dis <= '1' WHEN ctlr_wr_sosi.sync ='1' AND channel=g_start_channel ELSE '0'; END GENERATE;
  END GENERATE;

  p_reg : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      state <= s_flush;
    ELSIF rising_edge(clk) THEN
      state <= nxt_state;
    END IF;
  END PROCESS;
  
  p_state : PROCESS(state, dvr_wr_flush_en, dvr_done, dvr_en, dvr_wr_not_rd, flush_dis)
  BEGIN
    nxt_state <= state;
    ctlr_wr_flush_en <= '0';
    CASE state IS
      WHEN s_idle => 
        IF dvr_wr_flush_en='1' AND dvr_done='1' THEN
          ctlr_wr_flush_en <= '1';
          nxt_state <= s_flush;
        END IF;
      WHEN s_flush =>
        ctlr_wr_flush_en <= '1';
        IF dvr_en='1' AND dvr_wr_not_rd='1' THEN
          nxt_state <= s_stop;
        END IF;
      WHEN OTHERS => -- s_stop
        ctlr_wr_flush_en <= '1';
        IF flush_dis = '1' THEN
          nxt_state <= s_idle;
        END IF;
    END CASE;
  END PROCESS;
  
END str;

